@@ -266,8 +266,12 @@ static int _process_request_storage(mcp_parser_t *pr, size_t max) {
 }
 
 // common request with key: <cmd> <key> <args>
-static int _process_request_simple(mcp_parser_t *pr, const size_t max) {
+static int _process_request_simple(mcp_parser_t *pr, const int min, const int max) {
     _process_tokenize(pr, max);
+    if (pr->ntokens < min) {
+        P_DEBUG("%s: not enough tokens for simple request: %d\n", __func__, pr->ntokens);
+        return -1;
+    }
     pr->keytoken = 1; // second token is usually the key... stupid GAT.
 
     _process_request_key(pr);
@@ -351,7 +355,7 @@ int process_request(mcp_parser_t *pr, const char *command, size_t cmdlen) {
                     cmd = CMD_GET;
                     type = CMD_TYPE_GET;
                     token_max = 2; // don't chew through multigets.
-                    ret = _process_request_simple(pr, 2);
+                    ret = _process_request_simple(pr, 2, 2);
                 }
                 if (cm[1] == 'a' && cm[2] == 't') {
                     type = CMD_TYPE_GET;
@@ -375,13 +379,13 @@ int process_request(mcp_parser_t *pr, const char *command, size_t cmdlen) {
                 cmd = CMD_GETS;
                 type = CMD_TYPE_GET;
                 token_max = 2; // don't chew through multigets.
-                ret = _process_request_simple(pr, 2);
+                ret = _process_request_simple(pr, 2, 2);
             } else if (strncmp(cm, "incr", 4) == 0) {
                 cmd = CMD_INCR;
-                ret = _process_request_simple(pr, 4);
+                ret = _process_request_simple(pr, 3, 4);
             } else if (strncmp(cm, "decr", 4) == 0) {
                 cmd = CMD_DECR;
-                ret = _process_request_simple(pr, 4);
+                ret = _process_request_simple(pr, 3, 4);
             } else if (strncmp(cm, "gats", 4) == 0) {
                 cmd = CMD_GATS;
                 type = CMD_TYPE_GET;
@@ -393,7 +397,7 @@ int process_request(mcp_parser_t *pr, const char *command, size_t cmdlen) {
         case 5:
             if (strncmp(cm, "touch", 5) == 0) {
                 cmd = CMD_TOUCH;
-                ret = _process_request_simple(pr, 4);
+                ret = _process_request_simple(pr, 3, 4);
             } else if (strncmp(cm, "stats", 5) == 0) {
                 cmd = CMD_STATS;
                 // Don't process a key; fetch via arguments.
@@ -406,7 +410,7 @@ int process_request(mcp_parser_t *pr, const char *command, size_t cmdlen) {
         case 6:
             if (strncmp(cm, "delete", 6) == 0) {
                 cmd = CMD_DELETE;
-                ret = _process_request_simple(pr, 4);
+                ret = _process_request_simple(pr, 2, 4);
             } else if (strncmp(cm, "append", 6) == 0) {
                 cmd = CMD_APPEND;
                 ret = _process_request_storage(pr, token_max);
@@ -538,6 +542,7 @@ void mcp_request_attach(lua_State *L, mcp_request_t *rq, io_pending_proxy_t *p)
 // second argument is optional, for building set requests.
 // TODO: append the \r\n for the VAL?
 int mcplib_request(lua_State *L) {
+    LIBEVENT_THREAD *t = PROXY_GET_THR(L);
     size_t len = 0;
     size_t vlen = 0;
     mcp_parser_t pr = {0};
@@ -586,6 +591,13 @@ int mcplib_request(lua_State *L) {
             proxy_lua_error(L, "failed to allocate value memory for request object");
         }
         memcpy(rq->pr.vbuf, val, vlen);
+        // Note: Not enforcing the memory limit here is deliberate:
+        // - if we're over the memory limit, it'll get caught very soon after
+        // this, but we won't be causing some lua to bail mid-flight, which is
+        // more graceful to the end user.
+        pthread_mutex_lock(&t->proxy_limit_lock);
+        t->proxy_buffer_memory_used += rq->pr.vlen;
+        pthread_mutex_unlock(&t->proxy_limit_lock);
     }
 
     // rq is now created, parsed, and on the stack.
@@ -766,11 +778,15 @@ int mcplib_request_flag_token(lua_State *L) {
 }
 
 int mcplib_request_gc(lua_State *L) {
+    LIBEVENT_THREAD *t = PROXY_GET_THR(L);
     mcp_request_t *rq = luaL_checkudata(L, -1, "mcp.request");
     // During nread c->item is the malloc'ed buffer. not yet put into
     // rq->buf - this gets freed because we've also set c->item_malloced if
     // the connection closes before finishing nread.
     if (rq->pr.vbuf != NULL) {
+        pthread_mutex_lock(&t->proxy_limit_lock);
+        t->proxy_buffer_memory_used -= rq->pr.vlen;
+        pthread_mutex_unlock(&t->proxy_limit_lock);
         free(rq->pr.vbuf);
     }
 
