@@ -19,6 +19,8 @@
 /*
  * API functions
  */
+static void storage_finalize_cb(io_pending_t *pending);
+static void storage_return_cb(io_pending_t *pending);
 
 // re-cast an io_pending_t into this more descriptive structure.
 // the first few items _must_ match the original struct.
@@ -26,7 +28,10 @@ typedef struct _io_pending_storage_t {
     int io_queue_type;
     LIBEVENT_THREAD *thread;
     conn *c;
-    mc_resp *resp;            /* original struct ends here */
+    mc_resp *resp;
+    io_queue_cb return_cb;    // called on worker thread.
+    io_queue_cb finalize_cb;  // called back on the worker thread.
+                              /* original struct ends here */
     item *hdr_it;             /* original header item. */
     obj_io io_ctx;            /* embedded extstore IO header */
     unsigned int iovec_data;  /* specific index of data iovec */
@@ -119,12 +124,10 @@ void storage_stats(ADD_STAT add_stats, conn *c) {
 
 }
 
-
-// FIXME: This runs in the IO thread. to get better IO performance this should
-// simply mark the io wrapper with the return value and decrement wrapleft, if
-// zero redispatching. Still a bit of work being done in the side thread but
-// minimized at least.
-// TODO: wrap -> p?
+// This callback runs in the IO thread.
+// TODO: Some or all of this should move to the
+// io_pending's callback back in the worker thread.
+// It might make sense to keep the crc32c check here though.
 static void _storage_get_item_cb(void *e, obj_io *io, int ret) {
     // FIXME: assumes success
     io_pending_storage_t *p = (io_pending_storage_t *)io->data;
@@ -227,13 +230,7 @@ static void _storage_get_item_cb(void *e, obj_io *io, int ret) {
     p->active = false;
     //assert(c->io_wrapleft >= 0);
 
-    // All IO's have returned, lets re-attach this connection to our original
-    // thread.
-    io_queue_t *q = conn_io_queue_get(p->c, p->io_queue_type);
-    q->count--;
-    if (q->count == 0) {
-        redispatch_conn(c);
-    }
+    return_io_pending((io_pending_t *)p);
 }
 
 int storage_get_item(conn *c, item *it, mc_resp *resp) {
@@ -271,6 +268,9 @@ int storage_get_item(conn *c, item *it, mc_resp *resp) {
     p->miss = false;
     p->badcrc = false;
     p->noreply = c->noreply;
+    p->thread = c->thread;
+    p->return_cb = storage_return_cb;
+    p->finalize_cb = storage_finalize_cb;
     // io_pending owns the reference for this object now.
     p->hdr_it = it;
     p->resp = resp;
@@ -371,8 +371,12 @@ int storage_get_item(conn *c, item *it, mc_resp *resp) {
 void storage_submit_cb(io_queue_t *q) {
     // Don't need to do anything special for extstore.
     extstore_submit(q->ctx, q->stack_ctx);
+
+    // need to reset the stack for next use.
+    q->stack_ctx = NULL;
 }
 
+// Runs locally in worker thread.
 static void recache_or_free(io_pending_t *pending) {
     // re-cast to our specific struct.
     io_pending_storage_t *p = (io_pending_storage_t *)pending;
@@ -446,18 +450,17 @@ static void recache_or_free(io_pending_t *pending) {
     item_remove(p->hdr_it);
 }
 
-// Called after the IO is processed but before the response is transmitted.
-// TODO: stubbed with a reminder: should be able to move most of the extstore
-// callback code into this code instead, executing on worker thread instead of
-// IO thread.
-void storage_complete_cb(io_queue_t *q) {
-    // need to reset the stack for next use.
-    q->stack_ctx = NULL;
-    return;
+// Called after an IO has been returned to the worker thread.
+static void storage_return_cb(io_pending_t *pending) {
+    io_queue_t *q = conn_io_queue_get(pending->c, pending->io_queue_type);
+    q->count--;
+    if (q->count == 0) {
+        conn_worker_readd(pending->c);
+    }
 }
 
 // Called after responses have been transmitted. Need to free up related data.
-void storage_finalize_cb(io_pending_t *pending) {
+static void storage_finalize_cb(io_pending_t *pending) {
     recache_or_free(pending);
     io_pending_storage_t *p = (io_pending_storage_t *)pending;
     obj_io *io = &p->io_ctx;
@@ -571,8 +574,7 @@ static int storage_write(void *storage, const int clsid, const int item_age) {
 
 static pthread_t storage_write_tid;
 static pthread_mutex_t storage_write_plock;
-#define WRITE_SLEEP_MIN 500
-#define MIN_PAGES_FREE 3
+#define WRITE_SLEEP_MIN 200
 
 static void *storage_write_thread(void *arg) {
     void *storage = arg;
@@ -594,6 +596,10 @@ static void *storage_write_thread(void *arg) {
         int min_class = slabs_clsid(settings.ext_item_size);
         unsigned int global_pages = global_page_pool_size(NULL);
         bool do_sleep = true;
+        int target_pages = 0;
+        if (global_pages < settings.ext_global_pool_min) {
+            target_pages = settings.ext_global_pool_min - global_pages;
+        }
         counter++;
         if (to_sleep > settings.ext_max_sleep)
             to_sleep = settings.ext_max_sleep;
@@ -603,10 +609,8 @@ static void *storage_write_thread(void *arg) {
             bool mem_limit_reached = false;
             unsigned int chunks_free;
             int item_age;
+
             if (min_class > x || (backoff[x] && (counter % backoff[x] != 0))) {
-                // Long sleeps means we should retry classes sooner.
-                if (to_sleep > WRITE_SLEEP_MIN * 10)
-                    backoff[x] /= 2;
                 continue;
             }
 
@@ -614,12 +618,12 @@ static void *storage_write_thread(void *arg) {
             unsigned int chunks_perpage = 0;
             chunks_free = slabs_available_chunks(x, &mem_limit_reached,
                     &chunks_perpage);
-            unsigned int target = chunks_perpage * MIN_PAGES_FREE;
+            unsigned int target = chunks_perpage * target_pages;
 
             // storage_write() will fail and cut loop after filling write buffer.
             while (1) {
                 // if we are low on chunks and no spare, push out early.
-                if (chunks_free < target && global_pages <= settings.ext_global_pool_min) {
+                if (chunks_free < target) {
                     item_age = 0;
                 } else {
                     item_age = settings.ext_item_age;
@@ -627,6 +631,7 @@ static void *storage_write_thread(void *arg) {
                 if (storage_write(storage, x, item_age)) {
                     chunks_free++; // Allow stopping if we've done enough this loop
                     did_move = true;
+                    do_sleep = false;
                     if (to_sleep > WRITE_SLEEP_MIN)
                         to_sleep /= 2;
                 } else {
@@ -636,7 +641,7 @@ static void *storage_write_thread(void *arg) {
 
             if (!did_move) {
                 backoff[x]++;
-            } else if (backoff[x]) {
+            } else {
                 backoff[x] = 1;
             }
         }
@@ -644,6 +649,11 @@ static void *storage_write_thread(void *arg) {
         // flip lock so we can be paused or stopped
         pthread_mutex_unlock(&storage_write_plock);
         if (do_sleep) {
+            // Only do backoffs on other slab classes if we're actively
+            // flushing at least one class.
+            for (int x = 0; x < MAX_NUMBER_OF_SLAB_CLASSES; x++) {
+                backoff[x] = 1;
+            }
             usleep(to_sleep);
             to_sleep++;
         }
@@ -778,7 +788,7 @@ struct storage_compact_wrap {
 
 static void storage_compact_readback(void *storage, logger *l,
         bool drop_unread, char *readback_buf,
-        uint32_t page_id, uint64_t page_version, uint64_t read_size) {
+        uint32_t page_id, uint64_t page_version, uint32_t page_offset, uint64_t read_size) {
     uint64_t offset = 0;
     unsigned int rescues = 0;
     unsigned int lost = 0;
@@ -807,7 +817,8 @@ static void storage_compact_readback(void *storage, logger *l,
             if ((hdr_it->it_flags & ITEM_HDR) && !item_is_flushed(hdr_it) &&
                    (hdr_it->exptime == 0 || hdr_it->exptime > current_time)) {
                 hdr = (item_hdr *)ITEM_data(hdr_it);
-                if (hdr->page_id == page_id && hdr->page_version == page_version) {
+                if (hdr->page_id == page_id && hdr->page_version == page_version
+                        && hdr->offset == (int)offset + page_offset) {
                     // Item header is still completely valid.
                     extstore_delete(storage, page_id, page_version, 1, ntotal);
                     // drop inactive items.
@@ -893,7 +904,7 @@ static void *storage_compact_thread(void *arg) {
     bool compacting = false;
     uint64_t page_version = 0;
     uint64_t page_size = 0;
-    uint64_t page_offset = 0;
+    uint32_t page_offset = 0;
     uint32_t page_id = 0;
     bool drop_unread = false;
     char *readback_buf = NULL;
@@ -961,7 +972,8 @@ static void *storage_compact_thread(void *arg) {
                 LOGGER_LOG(l, LOG_SYSEVENTS, LOGGER_COMPACT_READ_START,
                         NULL, page_id, page_offset);
                 storage_compact_readback(storage, l, drop_unread,
-                        readback_buf, page_id, page_version, settings.ext_wbuf_size);
+                        readback_buf, page_id, page_version, page_offset,
+                        settings.ext_wbuf_size);
                 page_offset += settings.ext_wbuf_size;
                 wrap.done = false;
                 wrap.submitted = false;
@@ -1064,12 +1076,19 @@ struct extstore_conf_file *storage_conf_parse(char *arg, unsigned int page_size)
             multiplier = 1024 * 1024;
             multiplier *= 1024 * 1024 * 1024;
             break;
+        default:
+            fprintf(stderr, "must supply size to ext_path, ie: ext_path=/f/e:64m (M|G|T|P supported)\n");
+            goto error;
     }
     base_size = atoi(p);
     multiplier *= base_size;
     // page_count is nearest-but-not-larger-than pages * psize
     cf->page_count = multiplier / page_size;
     assert(page_size * cf->page_count <= multiplier);
+    if (cf->page_count == 0) {
+        fprintf(stderr, "supplied ext_path has zero size, cannot use\n");
+        goto error;
+    }
 
     // final token would be a default free bucket
     p = strtok_r(NULL, ",", &b);
